use actix::Addr;
use actix_web_actors::ws::WebsocketContext;
use serde::{Deserialize, Serialize};

use crate::terminal_size;
use actix::AsyncContext;
use alacritty_terminal::event::Notify;
use alacritty_terminal::event::{Event, EventListener};
use alacritty_terminal::event_loop::{EventLoop, Msg, Notifier};
use alacritty_terminal::sync::FairMutex;
use alacritty_terminal::term::cell::Cell;
use alacritty_terminal::term::{self, Term as Terminal};
use alacritty_terminal::{tty, Grid};

use elm_rs::{Elm, ElmDecode, ElmEncode};

// use tokio::sync::mpsc;

use std::collections::HashMap;
use std::sync::LazyLock;
use std::sync::{Arc, RwLock};
use std::thread;

type BackendType = crate::TryItBackend;
const MAX_TERMINALS: usize = 9;

static TERM_SESSION_BY_ACTOR: LazyLock<
    RwLock<HashMap<Addr<BackendType>, Arc<RwLock<TerminalSession>>>>,
> = LazyLock::new(|| RwLock::new(HashMap::new()));

#[derive(Deserialize, Debug, PartialEq, Elm, ElmEncode)]
#[serde(crate = "serde")]
pub enum TermUpMsg {
    RequestFullTermState,
    CloseTerm,
    RequestIncrementalTermStateUpdate,
    SendCharacter(String),
}

#[derive(Serialize, Debug, Elm, ElmDecode)]
pub enum TermDownMsg {
    TermUpdate(TerminalScreen),
    BackendTermStartFailure(String),
    TermNotStarted,
}

#[derive(Serialize, Debug, Elm, ElmDecode, PartialEq, Clone)]
pub struct Line {
    pub line: i32,
    pub content: String,
}

#[derive(Serialize, Debug, Elm, ElmDecode)]
pub struct Cursor {
    pub row: i32,
    pub col: usize,
}

#[derive(Serialize, Debug, Elm, ElmDecode)]
pub struct TerminalScreen {
    pub cols: usize,
    pub rows: usize,
    pub damaged_lines: Vec<Line>,
    // (line, col)
    pub cursor_loc: Cursor,
}

struct TerminalSession {
    term: Arc<FairMutex<Terminal<EventProxy>>>,
    old_content: String,
    // (line, col)
    old_cursor_pos: (i32, usize),
    /// Use to write to the terminal from the outside world.
    tx: Notifier,
    cols: u16,
    rows: u16,
}

#[derive(Clone)]
pub struct EventProxy {
    channel: std::sync::mpsc::Sender<Event>,
}

impl EventListener for EventProxy {
    fn send_event(&self, event: Event) {
        let _ = self.channel.send(event.clone());
    }
}

pub fn cleanup(actor_instance: &mut BackendType, ctx: &mut WebsocketContext<BackendType>) {
    if let Some(actor_id) = actor_instance.addr.clone() {
        if term_session_already_registered_in_hashmap(&actor_id) {
            let mut sessions = TERM_SESSION_BY_ACTOR.write().unwrap();
            if let Some(some_session) = sessions.get(&actor_id) {
                some_session
                    .read()
                    .unwrap()
                    .tx
                    .0
                    .send(Msg::Shutdown)
                    .unwrap();
                log::info!("Closed pty.");
            }

            sessions.remove(&actor_id);
            log::info!("Cleaned up terminal session for actor {:?}", actor_id);
        }
    }
}

pub fn msg_handler(
    msg: TermUpMsg,
    actor_instance: &mut BackendType,
    ctx: &mut WebsocketContext<BackendType>,
) {
    // only proceed if msg not CloseTerm
    if !(msg == TermUpMsg::CloseTerm) {
        if let Some(actor_id) = actor_instance.addr.clone() {
            if term_session_already_registered_in_hashmap(&actor_id) {
                let mut term_session = TERM_SESSION_BY_ACTOR.write().unwrap();
                if let Some(term_session) = term_session.get_mut(&actor_id) {
                    match msg {
                        TermUpMsg::RequestFullTermState => {
                            send_terminal_update_if_changed(&actor_id, term_session);
                        }
                        TermUpMsg::SendCharacter(c) => {
                            {
                                let term_session = term_session.write().unwrap();
                                term_session.tx.notify(c.to_string().into_bytes());
                                let mut term = term_session.term.lock();
                                term.scroll_display(alacritty_terminal::grid::Scroll::Bottom);
                            }
                            send_terminal_update_if_changed(&actor_id, term_session);
                        }
                        _ => {}
                    }
                }
            } else if get_session_count() >= MAX_TERMINALS {
                backend_term_start_error(
                    actor_instance,
                    ctx,
                    format!("Maximum number of {} open terminal sessions reached. Somebody must close at least one!", MAX_TERMINALS).to_string(),
                );
            } else {
                let (rows, cols) = (24, 80);

                let id = get_session_count() as u64;

                let pty_config = tty::Options {
                    shell: Some(tty::Shell::new("bash".to_string(), vec![])),
                    ..tty::Options::default()
                };
                let config = term::Config::default();
                let terminal_size = terminal_size::TerminalSize::new(rows, cols);
                let pty = match tty::new(&pty_config, terminal_size.into(), id) {
                    Ok(pty) => pty,
                    Err(msg) => {
                        backend_term_start_error(
                            actor_instance,
                            ctx,
                            format!("Failed to create pty: {}", msg),
                        );
                        return;
                    }
                };
                let (event_sender, event_receiver) = std::sync::mpsc::channel();
                let event_proxy = EventProxy {
                    channel: event_sender,
                };
                let term = Terminal::new::<terminal_size::TerminalSize>(
                    config,
                    &terminal_size,
                    event_proxy.clone(),
                );
                let term = Arc::new(FairMutex::new(term));
                let pty_event_loop =
                    match EventLoop::new(term.clone(), event_proxy, pty, false, false) {
                        Ok(loop_instance) => loop_instance,
                        Err(err) => {
                            backend_term_start_error(
                                actor_instance,
                                ctx,
                                format!("EventLoop::new failed: {:?}", err),
                            );
                            return;
                        }
                    };
                let notifier = Notifier(pty_event_loop.channel());
                let term_clone = term.clone();
                pty_event_loop.spawn();

                let actor_id_thread_clone = actor_id.clone();
                let term_thread_clone = term.clone();
                let addr_thread_clone = ctx.address().clone();
                ctx.spawn(actix::fut::wrap_future(async move {
                    // Start a new thread
                    thread::spawn(move || loop {
                        if let Ok(event) = event_receiver.recv() {
                            let mut sessions = TERM_SESSION_BY_ACTOR.write().unwrap();
                            match event {
                                Event::Exit => {
                                    sessions.remove(&actor_id_thread_clone);
                                    break;
                                }
                                _ => {
                                    if let Some(term_session) =
                                        sessions.get_mut(&actor_id_thread_clone)
                                    {
                                        send_terminal_update_if_changed(
                                            &addr_thread_clone,
                                            term_session,
                                        );
                                    }
                                }
                            }
                        };
                    });
                }));

                let terminal_session = TerminalSession {
                    term: term.clone(),
                    old_content: String::new(),
                    old_cursor_pos: (0, 0),
                    tx: notifier,
                    cols,
                    rows,
                };
                TERM_SESSION_BY_ACTOR
                    .write()
                    .unwrap()
                    .insert(actor_id, Arc::new(RwLock::new(terminal_session)));
            }
        }
    }
}

fn backend_term_start_error(
    actor_instance: &mut BackendType,
    ctx: &mut WebsocketContext<BackendType>,
    msg: String,
) {
    log::error!("{}", msg);
    let down_msg = crate::DownMsg::TermDownMsg(TermDownMsg::BackendTermStartFailure(msg));
    actor_instance.send_down_msg(down_msg, ctx);
}
fn get_session_count() -> usize {
    let sessions_map = TERM_SESSION_BY_ACTOR.read().unwrap();
    sessions_map.len()
}
fn term_session_already_registered_in_hashmap(actor_id: &Addr<BackendType>) -> bool {
    let sessions_map = TERM_SESSION_BY_ACTOR.read().unwrap();
    sessions_map.contains_key(actor_id)
}

fn full_term_state(
    terminal_session: &TerminalSession,
) -> (String, alacritty_terminal::index::Point) {
    let (rows, cols) = (terminal_session.rows, terminal_session.cols);
    let term = terminal_session.term.lock();
    let grid = term.grid().clone();
    let cursor = grid.cursor.point;

    (term_grid_to_string(&grid, rows, cols), cursor)
}

fn term_grid_to_string(grid: &Grid<Cell>, rows: u16, cols: u16) -> String {
    let mut term_content = String::with_capacity((rows * cols) as usize);

    // Populate string from grid
    for indexed in grid.display_iter() {
        let x = indexed.point.column.0;
        let y = indexed.point.line.0 as usize;
        if y < rows as usize && x < cols as usize {
            term_content.push(indexed.c);
        }
    }
    term_content
}

fn split_term_content_by_line(term_content: &str, cols: u16) -> Vec<Line> {
    let mut term_content_by_line: Vec<Line> = vec![];
    let term_content_chars: Vec<char> = term_content.chars().collect();
    for (index, chunk) in term_content_chars.chunks(cols as usize).enumerate() {
        let content = chunk.iter().collect::<String>();
        let line = Line {
            line: index as i32,
            content,
        };
        term_content_by_line.push(line);
    }
    term_content_by_line
}

fn send_terminal_update_if_changed(
    actor_instance: &Addr<BackendType>,
    term_session: &mut Arc<RwLock<TerminalSession>>,
) {
    let mut term_session = term_session.write().unwrap();

    let (new_content, new_cursor) = full_term_state(&term_session);
    let new_term_content_by_line = split_term_content_by_line(&new_content, term_session.cols);

    let term_content_len_has_changed = term_session.old_content.len() != new_content.len();
    let damaged_lines = if term_content_len_has_changed {
        new_term_content_by_line
    } else {
        let old_term_content_by_line =
            split_term_content_by_line(&term_session.old_content, term_session.cols);
        new_term_content_by_line
            .iter()
            .zip(old_term_content_by_line.iter())
            .filter(|(a, b)| a != b)
            .map(|(a, b)| (a.clone())) // Extract values from references
            .collect()
    };
    let down_msg = crate::DownMsg::TermDownMsg(TermDownMsg::TermUpdate(TerminalScreen {
        rows: term_session.rows as usize,
        cols: term_session.cols as usize,
        damaged_lines,
        cursor_loc: Cursor {
            row: new_cursor.line.0,
            col: new_cursor.column.0,
        },
    }));
    actor_instance.do_send(down_msg);
    term_session.old_content = new_content;
}
