module Page.Term exposing
    ( Model
    , Msg
    , view
    , init
    , update
    , subscriptions
    )
import Element exposing (Element)
import Websockets
import Ports
import Json.Decode as Decode
import Json.Encode as Encode exposing (Value)
import Html.Attributes exposing (placeholder)
import Browser.Events exposing (onKeyDown, onKeyUp)

import EncodeDecode

import List.Extra
import Element.Font
import Element.Input
import Element.Background
import Debug
import Element.Border
import String exposing (pad)

type alias RawKey = String

type Key
  = Character RawKey Char
  | Control RawKey String
  | Invalid

prependKey : Key -> List Key -> List Key
prependKey key keys =
    case key of
        Invalid ->
            keys
        _ ->
            if List.any (sameRawKey key) keys then
                keys
            else
                key :: keys

removeKey : Key -> List Key -> List Key
removeKey key keys =
    List.filter (not << sameRawKey key) keys

sameRawKey : Key -> Key -> Bool
sameRawKey key1 key2 =
    case (key1, key2) of
        (Character raw1 _, Character raw2 _) ->
            raw1 == raw2

        (Control raw1 _, Control raw2 _) ->
            raw1 == raw2

        _ ->
            False


keyDecoder : Decode.Decoder Key
keyDecoder =
    Decode.map2
        toKey
        (Decode.field "key" Decode.string)
        (Decode.field "code" Decode.string)

toKey : String -> String -> Key
toKey key code =
    case key of
        "Tab" -> Control code "Tab"
        _ ->
            case String.uncons key of
                Just (char, "") -> Character code char
                _ -> Control code key


isLowercaseAlpha : Char -> Bool
isLowercaseAlpha c =
    charIsBetweenInclusive c 'a' 'z'


processForCtrlChar : Char -> Char
processForCtrlChar c  =
    if isLowercaseAlpha c then
        Char.fromCode (Char.toCode c - Char.toCode 'a' + 1)
    else if charIsBetweenInclusive c '[' '_' then
        Char.fromCode (Char.toCode c - Char.toCode '[')
    else
        c


charIsBetweenInclusive : Char -> Char -> Char -> Bool
charIsBetweenInclusive c loChar hiChar =
    let
        cCode = Char.toCode c
        loCode = Char.toCode loChar
        hiCode = Char.toCode hiChar
    in
    cCode >= loCode && cCode <= hiCode

controlInKeyList : List Key -> Bool
controlInKeyList list =
    case list of
        [] -> False
        x :: xs ->
            case x of
                (Control _ "Control") -> True
                _ -> controlInKeyList xs


keyToString : List Key -> Maybe String
keyToString keyList =
    case keyList of
        (Character _ c) :: _ ->
            if (controlInKeyList keyList) then
                Just <| String.fromChar <| processForCtrlChar c
            else
                Just <| String.fromChar c
        (Control _ "Enter") :: _       -> Just "\r"
        (Control _ "Tab") :: _         -> Just "\t"
        (Control _ "Escape") :: _      -> Just "\u{001B}"
        (Control _ "Backspace") :: _   -> Just "\u{0008}"
        (Control _ "ArrowUp") :: _     -> Just "\u{001B}[A"
        (Control _ "ArrowDown") :: _   -> Just "\u{001B}[B"
        (Control _ "ArrowLeft") :: _   -> Just "\u{001B}[D"
        (Control _ "ArrowRight") :: _  -> Just "\u{001B}[C"

        _ -> Nothing


init : () -> (Model, Cmd Msg)
init flags =
    let
        -- Initial model with default terminal screen
        initialLine =
            { line = 0
            , content = "Spinning up container..."
            }
        initialCursor =
            { row = 0
            , col = 0
            }
        initialModel =
            { termState =
                { cols = String.length initialLine.content
                , rows = 1
                , lines = [initialLine]
                , cursorLoc = initialCursor
                }
            , heldKeys = []
            , err = Nothing
            }

        upMsg = EncodeDecode.TermUpMsg
                <| EncodeDecode.RequestFullTermState
        cmd =
            Ports.socketSend <| EncodeDecode.upMsgEncoder upMsg
    in
    (initialModel, cmd)

subscriptions : Model -> Sub Msg
subscriptions model =
    let
        ws = Ports.socketOnEvent
                (Websockets.EventHandlers
                    (SocketOpened)
                    (\_ -> NoOp)
                    (\_ -> NoOp)
                    (\message ->
                        case Decode.decodeString EncodeDecode.downMsgDecoder message.data of
                            Ok msg -> DownMsg msg
                            Err err -> DecodeError err
                    )
                    (\_ -> NoOp)
                )
        keyDownSubscription = onKeyDown (Decode.map KeyDown keyDecoder)
        keyUpSubscription   = onKeyUp   (Decode.map KeyUp keyDecoder)
        in
            Sub.batch [ ws
                      , keyDownSubscription
                      , keyUpSubscription
                      ]

makeGridWithNewlines : String -> Int -> String
makeGridWithNewlines string cols =
    let
        rows =
            string
                |> String.toList
                |> List.indexedMap Tuple.pair
                |> groupByRow cols
                |> List.map (List.map Tuple.second) -- Extract characters
                |> List.map String.fromList         -- Convert rows to strings
    in
        String.join "\n" rows


groupByRow : Int -> List (Int, Char) -> List (List (Int, Char))
groupByRow cols list =
    case list of
        [] ->
            []

        _ ->
            let
                (row, rest) =
                    List.Extra.splitAt cols list
            in
            row :: groupByRow cols rest

type alias TermState =
    { cols : Int
    , rows : Int
    , lines : List (EncodeDecode.Line)
    , cursorLoc : EncodeDecode.Cursor
    }

type alias Model = {
    termState : TermState,
    heldKeys : List Key,
    err : Maybe String
    }

type Msg
    = DownMsg EncodeDecode.DownMsg
    | DecodeError Decode.Error
    | GreetWidgetText String
    | SocketOpened Websockets.WebsocketOpened
    | KeyDown Key
    | KeyUp Key
    | NoOp

updateTermState : EncodeDecode.TerminalScreen -> TermState -> TermState
updateTermState termScreenUpdate currTermState =
    let
        fullUpdate = termScreenUpdate.rows == (List.length termScreenUpdate.damagedLines)
        newLines = case fullUpdate of
            True -> termScreenUpdate.damagedLines
            False -> updateLines currTermState.lines termScreenUpdate.damagedLines
    in
        { cols = termScreenUpdate.cols
        , rows = termScreenUpdate.rows
        , lines = newLines
        , cursorLoc = termScreenUpdate.cursorLoc
        }

updateLines : List EncodeDecode.Line -> List EncodeDecode.Line -> List EncodeDecode.Line
updateLines currLines damagedLines =
    List.map
        (\oldLine ->
            case List.filter (\damaged -> damaged.line == oldLine.line) damagedLines of
                [] ->
                    oldLine -- No replacement, keep the old line

                damaged :: _ ->
                    damaged -- Replace with the first matching damaged line
        )
        currLines

update : Msg -> Model -> (Model, Cmd Msg)
update msg model =
    case msg of
        DownMsg (EncodeDecode.TermDownMsg (EncodeDecode.TermUpdate terminalScreen)) ->
            ({ model | termState = updateTermState terminalScreen model.termState }, Cmd.none)
        DownMsg (EncodeDecode.TermDownMsg (EncodeDecode.BackendTermStartFailure err_msg)) ->
            ({ model | err = Just err_msg }, Cmd.none)
        SocketOpened _ ->
            let
                upMsg = EncodeDecode.TermUpMsg
                     <| EncodeDecode.RequestFullTermState
                cmd =
                    Ports.socketSend <| EncodeDecode.upMsgEncoder upMsg
            in
            (model, cmd)
        KeyDown key ->
            let
                newHeldKeys = prependKey key model.heldKeys
                processedChar = keyToString newHeldKeys
                cmd = case processedChar of
                        Just c ->
                            let
                                upMsg =
                                    EncodeDecode.TermUpMsg     <|
                                    EncodeDecode.SendCharacter <|
                                    c
                            in
                                Ports.socketSend <| EncodeDecode.upMsgEncoder upMsg
                        Nothing -> Cmd.none
            in
                ({model | heldKeys = newHeldKeys}, cmd)
        KeyUp key ->
            ({model | heldKeys = removeKey key model.heldKeys}, Cmd.none)
        _ -> (model, Cmd.none)

termFont = Element.Font.family
    [ Element.Font.typeface "Courier Prime"
    , Element.Font.monospace
    ]

lightOrangeBackground = Element.Background.color (Element.rgb255 255 87 51)
lightGreenBackground = Element.Background.color (Element.rgb255 175 225 175)

view : Model -> Element Msg
view model =
    let
        termContentString = case model.err of
            Nothing -> String.join "\n" (List.map .content model.termState.lines)
            Just err -> err
        useNotes =
            Element.column
                -- light orange background
                -- rounded corners
                [ lightGreenBackground
                , Element.padding 10
                , Element.spacing 5
                , Element.width (Element.fill)
                , Element.Border.rounded 10
                , Element.Font.color (Element.rgb255 255 255 255)
                ]
                [
                  Element.paragraph [] [Element.text <| "Written in Elm. Works well in Chrome. Your mileage with other browsers may vary!"],
                  Element.paragraph [] [Element.text <| "Container is nix based, so install programs with `nix-shell -p` vim htop for example."],
                  Element.paragraph [] [Element.text <| "Cursor and line higlighting not yet implmented.\n"],
                  Element.paragraph [] [Element.text <| "Container stats: 4 cores, 6gb memory"]
                ]
        cursorString =
            let
                cursorRow = model.termState.cursorLoc.row
                cursorCol = model.termState.cursorLoc.col

                paddingLeft = String.repeat cursorCol " "
                cursorChar = "█" -- Cursor box character
                paddingRight = String.repeat (model.termState.cols - cursorCol - 1) " "
                activeRow = paddingLeft ++ cursorChar ++ paddingRight

                standardRow = (String.repeat model.termState.cols " ")

                paddingAbove = String.repeat cursorRow standardRow
                paddingBelow = String.repeat (model.termState.rows - cursorRow - 1) standardRow

                stringBeforeNewlines = paddingAbove ++ activeRow ++ paddingBelow
            in
                case model.err of
                    Nothing ->
                        makeGridWithNewlines stringBeforeNewlines model.termState.cols
                    Just _ -> ""
        cursorLayer =
            Element.el
                [ termFont
                , Element.width (Element.fill |> Element.maximum 770 |> Element.minimum 770)
                , Element.height (Element.fill |> Element.maximum 390 |> Element.minimum 390)
                , Element.Font.color (Element.rgba255 0 0 0 0.3) -- White cursor
                ]
                (cursorString |> Element.text)


        mainTermLayer =
            Element.el
                [ termFont
                , Element.width (Element.fill |> Element.maximum 770 |> Element.minimum 770)
                , Element.height (Element.fill |> Element.maximum 390 |> Element.minimum 390)
                ]
                (termContentString |> Element.text)
        keyDebugger =
            Element.el
                [ lightOrangeBackground
                , Element.Border.rounded 10
                , Element.width (Element.fill)
                , Element.Font.color (Element.rgb255 255 255 255)
                , Element.padding 10
                ]
                (Element.text <| ("Debug keys: " ++ keysToText model.heldKeys))
    in
        Element.column
            [ Element.spacing 20
            , Element.alignTop
            , Element.alignLeft
            , Element.Font.size 16
            ]
            [ useNotes
            , Element.el
                [Element.inFront cursorLayer]
                (mainTermLayer)
            , keyDebugger
            ]

keysToText : List Key -> String
keysToText keys =
    keys
        |> List.map keyToStringDebug
        |> String.join ""

keyToStringDebug : Key -> String
keyToStringDebug key =
    case key of
        Character _ c ->
            if c == ' ' then
                "⊔"
            else
                String.fromChar c

        Control _ str ->
            "<" ++ str ++ ">"
        _ -> ""
